open Printf
open Globals
open Time
open Vec
open Event
open Util


module EV = struct
  let pwc n x0 evs = V.comp1 id (V.CompPWC_n (n, x0, evs))

  let dpwl n x0 devs =
    let dx = pwc n 0.0 devs
    in V.accumulate (+.) x0 dx

  let pwl n x0 evs = 
    let f (Event (x1, t1)) (Event (x2, t2)) = 
      ((x2 -. x1) /. (float (t2 - t1))) @@ t1
    in dpwl n x0 (V.map_by2 f (x0 @@ 0) evs)

  let vfold f x0 vevs = 
    let nevs = V.length vevs
    and starts = V.map E.t vevs
    and ends = V.map (fun (Event (x, t)) -> t + (V.length x)) vevs
    in let live i t = (t >= starts.(i) && t < ends.(i))
    in let len = V.foldl max 0 ends
    in let t0 = ref 0
       and t1 = ref 0
       and ax = ref x0

    in let ft t = (
      while (!t0 < nevs &&   ends.(!t0) <= t) do incr t0 done;
      while (!t1 < nevs && starts.(!t1) <= t) do incr t1 done;
      ax := x0;
      for i = !t0 to (!t1 - 1) do
        if (live i t) 
        then let vi = E.v (vevs.(i))
             and j = t - starts.(i)
             in let e = vi.(j)
             in ax := f !ax e
        else ()
      done;
      !ax
     )
    in V.table len ft

end;;
