open Printf
open Globals
open Time
open Event
open Util

type 'a vec = 'a array

module V = struct
  let copy x = Array.copy x
  let const n xi = Array.make n xi
  let table n f = Array.init n f
  let vec x0 = const 1 x0

  let length = Array.length

  let take n x = Array.sub  x 0                n
  let drop n x = Array.sub  x n                ((length x) - n)
  let rtake n x = Array.sub x ((length x) - n) n
  let rdrop n x = Array.sub x 0                ((length x) - n)
  let sub n len x = 
(*    (eprintf "n %d, len %d\n" n len; *)
     Array.sub x n len

  let join = Array.append
  let delay n x0 x = join (const n x0) x

  let peek n x = Array.get x n
  let poke n e x = Array.set x n e
  let fpoke n e x = 
    let x' = copy x 
    in Array.set x' n e;
    x'
  let first x = (peek 0 x)
  let last  x = (peek ((length x) - 1) x)

  let peekv nv x = table (length nv) (fun i -> peek i x)
  let fpokev nv ev x = 
    let x' = copy x
    in Array.iteri (fun i ni -> Array.set x ni (peek i ev)) nv;
    x'

  let map f x = Array.map f x
  let mapi f x = Array.mapi f x
  let foldl f y0 x = Array.fold_left  f y0 x
  let foldr f x y0 = Array.fold_right f x y0
  let filter f x = 
    let rec list_filter = function 
        (_, [])   -> []
      | (f, h::t) -> if (f h) then h::(list_filter (f, t))
                              else    (list_filter (f, t))
    in Array.of_list (list_filter (f, Array.to_list x))
  let find f x = 
    let rec list_find_rev = function
        (_, ax, _, [])   -> ax
      | (i, ax, f, h::t) -> 
          list_find_rev (i+1, (if (f h) then (i::ax) else ax), f, t)
    in Array.of_list (List.rev (list_find_rev (0, [], f, Array.to_list x)))
  let compress b x = peekv (find id b) x

  let ravel2 x = foldl join [||] x

  let unzip x = (map fst x, map snd x)
  let unzip3 x = 
    (map (fun (e, _, _) -> e) x,
     map (fun (_, e, _) -> e) x,
     map (fun (_, _, e) -> e) x)
  let unzip4 x = 
    (map (fun (e, _, _, _) -> e) x,
     map (fun (_, e, _, _) -> e) x,
     map (fun (_, _, e, _) -> e) x,
     map (fun (_, _, _, e) -> e) x)

  let scale a x = map (fun x -> a*.x) x

(** conversion to string **)

  let to_string elt_to_string x = 
    let sx = (foldl (fun str xi -> str ^ " " ^ (elt_to_string xi)) "" x)
    in let snippage = if ((length x) > 0) then 1 else 0
    in "[" ^ (String.sub sx snippage ((String.length sx) - snippage)) ^ "]"
  let to_text elt_to_string spacer x = 
    foldl (fun str xi -> str ^ (elt_to_string xi) ^ spacer) "" x
  let matrix_to_text elt_to_string x = 
    foldl (fun str xrow -> str ^ (to_text elt_to_string " " xrow) ^ "\n") "" x

  let   int_to_string x = to_string string_of_int x
  let float_to_string x = to_string string_of_float x
  let   int_to_text x = to_text string_of_int "\n" x
  let float_to_text x = to_text string_of_float "\n" x
  let float_matrix_to_text x = matrix_to_text string_of_float x
  let istring =   int_to_string
  let fstring = float_to_string
  let itext   =   int_to_text
  let ftext   = float_to_text
  let ftext2  = float_matrix_to_text

  let print_text_float x = 
    let rec print_text_float' x i imax =
      (print_string (string_of_float (peek i x));
       print_string "\n";
       if i < imax-1 then (print_text_float' x (i+1) imax)
       else ()
      )
    in print_text_float' x 0 (length x)

(** sequences **)

  let icount_by from by n = table n (fun i -> (from + by*i)) 
  let icount from n = icount_by from 1 n

  let count_by from by n = table n (fun i -> (from +. by *. (float i)))
  let count from n = count_by from 1.0 n
  let count_n_to from to' n =           (* up to and to *)
    count_by from ((to' -. from) /. (float (n-1))) n
  let count_n_upto from upto n =        (* up to but not to *)
    count_by from ((upto -. from) /. (float n)) n
  let count_by_upto from by upto =      (* up to but not to *)
    let k = (upto -. from) /. by
    in let nk = int_of_float k
    in let n = nk + if ((k -. (float nk)) < floatfuzz) then 0 else 1
    in count_by from by n

(** comprehensions **)
  type 'a compsrc = 
      CompIn of 'a vec                       (* (xi) in x *)
    | CompPWC of 'a * 'a event vec           (* (xi) pwc x0 evs *)
    | CompPWC_n of time * 'a * 'a event vec  (* (xi) pwc len x0 evs *)
  type 'a compiter =
      CompIter  of 'a * time                 (* (i, ax) iter x0 len *)

  let cs_len cs = match cs with
    (CompIn x)            -> (Some (length x))
  | (CompPWC_n (n, _, _)) -> (Some n)
  | (CompPWC _)           ->  None

  let ci_len ci = match ci with
    (CompIter  (_, n))    -> if (n = -1) then None else (Some n)

  let rec comp1 f cs = match cs with 
    (CompIn   x)       -> map f x
  | CompPWC (x0, evs) ->
      comp1 f (CompPWC_n (E.t (last evs), x0, evs))
  | (CompPWC_n (n, x0, evs)) ->
      let xi = ref x0
      and ie = ref (-1)
      and et = ref time_0
      and ev = ref x0
      in let loop = fun i -> 
        (while (i >= !et) 
         do (xi := !ev;
             incr ie; 
             if (!ie = (length evs))
             then et := time_max
             else let e = (peek !ie evs) 
                  in (et := E.t e; ev := E.v e)
            )
         done;
         !xi
        )
      in table n loop

  let rec comp1i f ci = match ci with 
    (CompIter (x0, n)) -> 
      let ax = ref x0
      and i = ref time_0
      in let loop = fun _ -> 
        let oldax = !ax in (ax := f !i !ax;  incr i;  oldax)
      in table n loop

  let unify = function 
      (Some x, Some y) -> if x=y then (Some x) else raise (Length (x, y))
    | (Some x, None  ) -> (Some x)
    | (None  , Some y) -> (Some y)
    | (None  , None  ) ->  None
  let force_some = function 
      (Some x) -> x 
    | _ -> raise Length_underspecified
  let length2 cs1 cs2 = 
    force_some (unify (cs_len cs1, cs_len cs2))
  let length2i ci cs = 
    force_some (unify (ci_len ci, cs_len cs))
  let lengthn lenlist = 
    force_some (List.fold_left (fun len cslen -> unify (len, cslen))
                  None lenlist)
  let fixlen len cs = match cs with
    (CompPWC (x0, evs)) -> (CompPWC_n (len, x0, evs))  
  | _ -> cs 

  let comp2 f cs1 cs2 = 
    let len = length2 cs1 cs2
    in let vx1 = comp1 id (fixlen len cs1)
       and vx2 = comp1 id (fixlen len cs2)
    in let loop i = f (peek i vx1) (peek i vx2)
    in table len loop

  let comp2i f ci cs = 
    let len = length2i ci cs
    in let vy = comp1 id (fixlen len cs)
    in let f' i ax = f i ax (peek i vy)
    in let (CompIter (x0, _)) = ci
    in comp1i f' (CompIter (x0, len))

  let comp3 f cs1 cs2 cs3 = 
    let len = lengthn [cs_len cs1; cs_len cs2; cs_len cs3]
    in let vx1 = comp1 id (fixlen len cs1)
       and vx2 = comp1 id (fixlen len cs2)
       and vx3 = comp1 id (fixlen len cs3)
    in let loop i = f (peek i vx1) (peek i vx2) (peek i vx3)
    in table len loop

  let comp3i f ci cs1 cs2 = 
    let len = lengthn [ci_len ci; cs_len cs1; cs_len cs2]
    in let vx1 = comp1 id (fixlen len cs1)
       and vx2 = comp1 id (fixlen len cs2)
    in let f' i ax = f i ax (peek i vx1) (peek i vx2)
    in let (CompIter (x0, _)) = ci
    in comp1i f' (CompIter (x0, len))

  let comp4i f ci cs1 cs2 cs3 = 
    let len = lengthn [ci_len ci; cs_len cs1; cs_len cs2; cs_len cs3]
    in let vx1 = comp1 id (fixlen len cs1)
       and vx2 = comp1 id (fixlen len cs2)
       and vx3 = comp1 id (fixlen len cs3)
    in let f' i ax = f i ax (peek i vx1) (peek i vx2) (peek i vx3)
    in let (CompIter (x0, _)) = ci
    in comp1i f' (CompIter (x0, len))

  let map2 f x y   = comp2 f (CompIn x) (CompIn y)
  let map3 f x y z = comp3 f (CompIn x) (CompIn y) (CompIn z)

  let map_by2 f x0 x =
    let f1 i = f (if (i=0) then x0 else (peek (i-1) x)) (peek i x)
    in table (length x) f1

  let iterate n f x0 = 
    comp1i (fun i ax -> f ax) (CompIter (x0, n))
  let accumulate f x0 x = 
    comp1i (fun i ax -> f ax (peek i x)) (CompIter (x0, length x))

end;;
